# -*- coding: utf-8 -*-

from docutils import nodes
import os
import re

class Emumerator(nodes.SparseNodeVisitor):
    def __init__(self, document, project):
        nodes.SparseNodeVisitor.__init__(self, document)
        self.project = project
        self.section_level = 0
        self.list_level = 0
        self.storage = {}
        self.refs = {}
        numbering = project['numbering'] if 'numbering' in project else None
        self.suffix = numbering['suffix'].decode('unicode-escape') if numbering and ('suffix' in numbering) else None

        self.figure_suffix = numbering['figure-suffix'].decode('unicode-escape') if numbering and ('figure-suffix' in numbering) else None
        if self.figure_suffix is None:
            self.figure_suffix = self.suffix

        self.table_suffix = numbering['table-suffix'].decode('unicode-escape') if numbering and ('table-suffix' in numbering) else None
        if self.table_suffix is None:
            self.table_suffix = self.suffix

        self.list_suffix = numbering['list-suffix'].decode('unicode-escape') if numbering and ('list-suffix' in numbering) else None
        if self.list_suffix is None:
            self.list_suffix = self.suffix
        self.space = numbering['space'].decode('unicode-escape') if numbering and ('space' in numbering) else ' '

    def __process_list_in_children(self, children):
        for child in children:
            if isinstance(child, nodes.enumerated_list):
                self.visit_enumerated_list(child)
            elif len(child.children):
                self.__process_list_in_children(child.children)

    def visit_section(self, node):
        self.section_level += 1
        n = number('heading' + str(self.section_level), self.project, self.storage)
        if n:
            for child in node.children:
                if isinstance(child, nodes.title):                    
                    child.children[0] = nodes.Text(n + self.space + child.astext())
                    break

    def depart_section(self, node):
        self.section_level -= 1

    def visit_figure(self, node):
        n = number('figure', self.project, self.storage)
        if n:
            numbered = True
            if node.hasattr('ids'):
                for node_id in node['ids']:
                    self.refs[node_id] = n
                    if node_id.startswith('unnumbered'):
                        numbered = False
            if numbered:
                suffix = self.figure_suffix if self.figure_suffix != None else ''
                for child in node.children:
                    if isinstance(child, nodes.caption):                    
                        child.children[0] = nodes.Text(n + suffix + self.space + child.astext())
                        break

    def visit_table(self, node):
        n = number('table', self.project, self.storage)
        if n:
            numbered = True
            if node.hasattr('ids'):
                for node_id in node['ids']:
                    self.refs[node_id] = n
                    if node_id.startswith('unnumbered'):
                        numbered = False
            if numbered:
                suffix = self.table_suffix if self.table_suffix != None else ''
                for child in node.children:
                    if isinstance(child, nodes.title):                    
                        child.children[0] = nodes.Text(n + suffix + self.space + child.astext())
                        break

    def visit_enumerated_list(self, node):
        children = []
        for child in node.children:
            if isinstance(child, nodes.list_item) and len(child.children):
                children.append(child)
        if not len(children):
            return
        suffix = self.list_suffix if self.list_suffix != None else ''
        self.list_level += 1
        n = number('list' + str(self.list_level), self.project, self.storage)
        if n:
            if node.hasattr('ids'):
                for node_id in node['ids']:
                    self.refs[node_id] = n
            newnode = nodes.table()
            newnode['classes'] = ['item-list']
            tgroup = nodes.tgroup(cols=2)
            newnode += tgroup
            tbody = nodes.tbody()
            tgroup += tbody
            i = 0
            for child in children:
                self.__process_list_in_children(child.children)
                trow = nodes.row()
                tbody += trow
                tentry1 = nodes.entry()
                trow += tentry1
                tentry1 += nodes.paragraph(text=n + suffix)
                tentry2 = nodes.entry()
                trow += tentry2
                tentry2 += child.children
                i += 1
                if i < len(children):
                    n = number('list' + str(self.list_level), self.project, self.storage)
        
            node.replace_self(newnode)
        self.list_level -= 1

class ReferenceUpdater(nodes.SparseNodeVisitor):
    def __init__(self, document, refs):
        nodes.SparseNodeVisitor.__init__(self, document)
        self.refs = refs

    def visit_reference(self, node):
        if node.hasattr('refid'):
            refid = node['refid']
            if refid in self.refs:
                node.children[0] = nodes.Text(self.refs[refid])


def common(text, dir, project):
    # Substitute fields by values
    if 'fields' in project:
        for field in project['fields']:
            text = re.sub(r'\#\#\#' + field + r'\#\#\#', str(project['fields'][field]), text, flags=re.I)
            text = re.sub(r'\#\{' + field + r'\}', str(project['fields'][field]), text, flags=re.I)
    # Make path to images absolute
    if 'images-root' in project:
        dir = os.path.abspath(project['images-root'])
    dir = dir.replace('\\', '/')
    text = re.sub(r'((figure|image)::\S*\s+)([\w\d\/\\-]+\.(jpg|jpeg|pdf|png|svg))', r'\1' + dir + r'/\3', text, flags=re.I|re.M)

    # Add extra EOL
    text += '\n\n'
    return text

def doctree(doctree, project):
    if 'numbering' in project:
        e = Emumerator(doctree, project)
        doctree.walkabout(e)
        ru = ReferenceUpdater(doctree, e.refs)
        doctree.walk(ru)
    return doctree

def html(text, dir, project):
    return common(text, dir, project)

def number(tag, project, storage):
    result = None
    if 'numbering' in project:
        numbering = project['numbering']
        if tag in numbering:
            template = numbering[tag]
            patterns = [r'\(((\w+\d*)(?:([\+\-\=])(\d+)?)?)\)',
                r'\[((\w+\d*)(?:([\+\-\=])(\d+)?)?)\]',]
            for pattern in patterns:
                m = re.search(pattern, template, re.I)
                while m:
                    var = m.group(2)
                    operation = m.group(3)
                    operand = m.group(4)
                    value = storage[var] if var in storage else 0
                    if operation == '+':
                        value += int(operand) if operand else 1
                    if operation == '-':
                        value -= int(operand) if operand else 1
                    elif operation == '=':
                        value = int(operand) if operand else 0
                    storage[var] = value
                    number = str(value) if pattern.startswith('\(') else ''
                    template = re.sub(pattern, number, template, 1)                
                    m = re.search(pattern, template, re.I)

            result = template

    return result

def pdf(text, dir, project):
    return common(text, dir, project)


# -*- coding: utf-8 -*-

import os
import re
from collections import namedtuple

import docutils
from docutils.parsers.rst import directives

from rst2pdf.createpdf import RstToPdf, add_extensions
from rst2pdf import pygments_code_block_directive

from doker import fileutils, preprocess, log

def html(files, project):
    return

def pdf(files, project, output):
    generated = []
    pdf = project['pdf'] if 'pdf' in project else None

    # Stylesheet processing
    stylesheets = []
    try:
        if pdf and ('stylesheet' in pdf):
            stylesheet_json = fileutils.stylesheet_to_json(pdf['stylesheet'])
            generated.append(stylesheet_json)
            stylesheets.append(stylesheet_json)
        if pdf and ('stylesheets' in pdf):
            for stylesheet_file in pdf['stylesheets']:
                stylesheet_json = fileutils.stylesheet_to_json(stylesheet_file)
                generated.append(stylesheet_json)
                stylesheets.append(stylesheet_json)
    except Exception:
        fileutils.remove(generated)
        raise

    # Revisions
    if 'revisions' in project:
        revisions_text  = '.. _unnumbered_revisions:\n'
        revisions_text += '.. list-table:: Revision History\n'
        revisions_text += '   :class: revisions-table\n'
        revisions_text += '   :header-rows: 1\n'
        revisions_text += '\n'
        revisions_text += '   * - Version\n'
        revisions_text += '     - Date\n'
        revisions_text += '     - Description\n'
        
        last_version = None
        last_date = None
        revisions = project['revisions']
        for revision in revisions:
            if isinstance(revision, dict):
                key = revision.keys()[0]
                ver = key
                date = '--'
                m = re.search(r'([\w\d.-:]+)\s+(?:\(|\[)([0-9\w\s./:]+)(?:\)|\])', key)
                if m:
                    ver = m.group(1)
                    date = m.group(2)
                if not last_version:
                    last_version = ver
                if not last_date:
                    last_date = date
                revisions_text += '   * - **' + str(ver) + '**\n'
                revisions_text += '     - ' + date + '\n'
                revisions_text += '     - .. class:: revision-list\n'
                revisions_text += '\n'
                changes = revision[key]
                for change in changes:
                    revisions_text += '       + ' + change + '\n'
                    revisions_text += '\n'

        if not 'fields' in project:
            project['fields'] = {}
        project['fields']['revisions'] = revisions_text
        if (not 'version' in project['fields']) and last_version:
            project['fields']['version'] = last_version
        if (not 'date' in project['fields']) and last_date:
            project['fields']['date'] = last_date

    # Text processing
    text = ''

    # Cover page
    if pdf and ('cover' in pdf):
        cover_file = pdf['cover']
        if not os.path.isfile(cover_file):
            log.error("Unable to open cover file: '%s'", cover_file)
            fileutils.remove(generated)
            raise IOError('Cover file error')
        with open(cover_file, 'r') as f:
            text += preprocess.pdf(f.read(), os.path.dirname(cover_file), project)

    # Main contents processing
    for file in files:
        with open(file['src'], 'r') as f:
            text += preprocess.pdf(f.read(), os.path.dirname(file['src']), project)

    # Output file name processing
    if pdf and 'output' in pdf:
        output = os.path.join(pdf['output'], output) if os.path.isdir(pdf['output']) else pdf['output']
        if not output.endswith('.pdf'):
            output += '.pdf'

    # Generating PDF
    options = namedtuple('Namespace', 'extensions')
    options.extensions = ['inkscape_r2p', 'vectorpdf_r2p']
    if pdf and ('toc' in pdf):
        if pdf['toc'] == 'dotted':
            options.extensions.append('dotted_toc')
    add_extensions(options)
    directives.register_directive('code-block', pygments_code_block_directive.code_block_directive)
    directives.register_directive('code', pygments_code_block_directive.code_block_directive)
    doctree = docutils.core.publish_doctree(text)
    doctree = preprocess.doctree(doctree, project)
    log.info("Generating '%s'", os.path.basename(output))
    try:
        RstToPdf(
            background_fit_mode='scale',
            breaklevel=pdf['break-level'] if pdf and ('break-level' in pdf) else 1,
            breakside=pdf['break-side'] if pdf and ('break-side' in pdf) else 'any',
            repeat_table_rows=pdf['repeate-table-rows'] if pdf and ('repeate-table-rows' in pdf) else True,
            smarty=str(pdf['smartquotes'] if pdf and ('smartquotes' in pdf) else 1),
            stylesheets=stylesheets, 
        ).createPdf(doctree=doctree, output=output)
    except Exception as err:
        log.error('PDF generating failed')
        fileutils.remove(generated)
        raise

    log.info("Post-processing '%s'", os.path.basename(output))
    try:
        import pkg_resources
        __version__ = pkg_resources.get_distribution('doker').version
    except Exception:
        __version__ = None
    creator = 'Doker'
    if __version__:
        creator += ' v' + __version__
    creator += ' \\(doker.org\\)'
    with open(output, 'rb') as f:
        text = f.read()
    with open(output, 'wb') as f:
        f.write(re.sub(r'/Creator \(\\\(unspecified\\\)\)', '/Creator (' + creator + ')' , text))

    # Temporary files removing
    fileutils.remove(generated)



# -*- coding: utf-8 -*-

import logging

logging.basicConfig(format='[%(levelname)s] %(message)s')
log = logging.getLogger('doker')
log.setLevel(logging.INFO)

def critical(msg, *args, **kwargs):
    return log.critical(msg, *args, **kwargs)

def error(msg, *args, **kwargs):
    return log.error(msg, *args, **kwargs)

def info(msg, *args, **kwargs):
    return log.info(msg, *args, **kwargs)

def warning(msg, *args, **kwargs):
    return log.warning(msg, *args, **kwargs)


# -*- coding: utf-8 -*-

import json
import re
import os
import yaml

from collections import OrderedDict
from doker import log

def to_key(file):
    key = re.sub(r'^\d+-', '', file) # Remove number at the beginning
    key = re.sub(r'\..*$', '', key)  # Remove suffix
    return key

def get_title(src):
    with open(src, 'r') as f:
        title = f.readline().strip()
    return title

def get_tree(dir, file_tree=None):
    if not file_tree:
        file_tree = OrderedDict()
    files = os.listdir(dir)
    if files:
        files.sort()
    for file in files:
        file_path = dir + '/' + file
        if os.path.isfile(file_path) and not file.endswith('.rst'):
            continue

        key = to_key(file)
        if os.path.isdir(file_path):
            file_tree[key] = {}
            get_tree(file_path, file_tree)
            if not file_tree[key]:
                file_tree.pop(key, None)
        else:
            file_tree[key] = file_path

    return file_tree

def remove(file_list):
    for file in file_list:
        log.info("Removing temporary '%s'", os.path.basename(file))
        os.remove(file)

def stylesheet_to_json(stylesheet_file):
    if not os.path.isfile(stylesheet_file):
        log.error("Unable to open stylesheet file: '%s'", stylesheet_file)
        raise IOError('Stylesheet file error')
    with open(stylesheet_file, 'r') as f:
        try:
            style = yaml.load(f)
        except yaml.YAMLError as err:
            log.error("Parsing YAML style file failed: '%s'", err)
            raise

    stylesheet_name = 'tmp-stylesheet'
    i = 0
    while True:
        stylesheet_json = stylesheet_name + str(i) + '.json'
        if not os.path.isfile(stylesheet_json):
            break
        i += 1
    log.info("Generating temporary '%s'", stylesheet_json)
    with open(stylesheet_json, 'w') as f:
        f.write(json.dumps(style))

    return stylesheet_json

def to_list(file_tree):
  obj_list = []
  tree_branch(file_tree, None, 0, obj_list)
  return obj_list

def to_tree(file_list):
    file_tree = {}
    for file in file_list:
        file_tree[to_key(file)] = file
    return file_tree

def tree_branch(file_tree, dir, level, obj_list):
    children = []
    for k in file_tree.keys():
        v = file_tree[k]
        obj = {'path' : (dir + '/' + k) if dir else k, 'level': level + 1 }
        if isinstance(v, dict):
            obj['src'] = v['index']
            obj['title'] = get_title(obj['src'])
            obj_list.append(obj)
            obj.children = tree_branch(v, obj['path'], obj['level'], obj_list)
        elif k != 'index':
            obj['src'] = v
            obj['title'] = get_title(obj['src'])
            obj_list.append(obj)
            children.append(obj)

    return children


#!/usr/bin/env python

import re
from os import path 
from setuptools import find_packages, setup

current_dir = path.abspath(path.dirname(__file__))

with open(path.join(current_dir, 'CHANGELOG.rst'), 'r') as f:
    version_string = f.readline().strip()
m = re.search(r'(\d+\.\d+\.\d+)', version_string)
__version__ = m.group(1) if m else '0.0.0'

setup(
    name='doker',
    version=__version__,
    packages=find_packages(),
    author='Doker Authors',
    description="Rich PDF documents creating tool",
    url='https://github.com/doker-project/doker',
    long_description=open('README.rst').read(),
    entry_points={
        'console_scripts': [
            'doker = doker.main:main',
        ]},
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={
        '': ['CHANGELOG.rst', 'LICENSE']
    },
    install_requires = [
        'docutils',
        'pyyaml',
        'rst2pdf',
    ],
)

